# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 11:17:22 2016

@author: mwojc
"""

if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file
