#!/usr/bin/env python
# -*- coding: utf-8 -*-
#try:
from setuptools import setup, findall
#except ImportError:
    #from distutils.core import setup


if __name__ == "__main__":
    setup(name              = 'fempy',
          version           = '0.3.1',
          description       = 'Finite element method in python',
          long_description  = open('README.md', 'r').read(),
          long_description_content_type='text/markdown',
          keywords          = ['FEM, fem, finite element method'],
          author            = 'Marek Wojciechowski',
          author_email      = 'mwojc@p.lodz.pl',
          url               = 'http://fempy.org',
          license           = 'LGPL-3',
          platforms         = 'Posix, Windows',
          package_dir       = {'fempy': 'fempy',
                               'fempy.geometry': 'fempy/geometry',
                               'fempy.domain': 'fempy/domain',
                               'fempy.fields': 'fempy/fields',
                               'fempy.models': 'fempy/models',
                               'fempy.solvers': 'fempy/solvers',
                               'fempy.tests': 'tests',
                               'fempy.tests.geometry': 'tests/geometry',
                               'fempy.tests.domain': 'tests/domain',
                               'fempy.tests.fields': 'tests/fields',
                               'fempy.tests.models': 'tests/models',
                               'fempy.tests.solvers': 'tests/solvers',
                               'fempy.examples': 'examples'},
          packages          = ['fempy',
                               'fempy.geometry',
                               'fempy.domain',
                               'fempy.fields',
                               'fempy.models',
                               'fempy.solvers',
                               'fempy.tests',
                               'fempy.tests.geometry',
                               'fempy.tests.domain',
                               'fempy.tests.fields',
                               'fempy.tests.models',
                               'fempy.tests.solvers',
                               'fempy.examples'],
          package_data      = {'fempy.tests': ['data/*',
                                               'data/gmsh/*',
                                               'data/elastic/*',
                                               'data/heat/*']},
          #packages = find_packages(),
          #package_data      = {'tests': ['tests/data/*']},
          install_requires  = ['numpy', 'scipy'],
          classifiers       = ['Development Status :: 4 - Beta',
                               'Intended Audience :: Education',
                               'Intended Audience :: Science/Research',
                               'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
                               'Operating System :: POSIX :: Linux',
                               'Operating System :: Microsoft :: Windows',
                               'Operating System :: MacOS :: MacOS X',
                               'Programming Language :: C',
                               'Programming Language :: C++',
                               'Programming Language :: Python :: Implementation :: CPython',
                               'Topic :: Scientific/Engineering'],
          )
